# Copyright Contributors to the Opensynth-energy Project.
# SPDX-License-Identifier: Apache-2.0

import ast
from pathlib import Path

import numpy as np
import pandas as pd
import pytorch_lightning as pl
import torch
from torch.utils.data import DataLoader, Dataset

RANDOM_STATE = 0
g = torch.Generator()
g.manual_seed(RANDOM_STATE)


class LCLData(Dataset):
    """
    Low CarbonLondon Dataset
    """

    def __init__(
        self,
        data_path: Path,
        stats_path: Path,
        n_samples: int,
    ):
        """
        Args:
            data_path (Path): Data path
            stats_path (Path): Stats path. Note when loading
            evaluation dataset, the stats path point to the
            stats of training data, rather than evaluation data
            to avoid data leakage!
            n_samples (int): Number of samples to load
        """
        self.df = pd.read_csv(f"{data_path}/lcl_data.csv")
        self.df_stats = pd.read_csv(f"{stats_path}/mean_std.csv")

        # Parse stats
        self.feature_mean = self.df_stats["mean"].values[0]
        self.feature_std = self.df_stats["stdev"].values[0]

        # Resample Dataset
        self.n_samples = n_samples
        self.df = self.df.sample(
            self.n_samples, random_state=RANDOM_STATE
        ).reset_index(drop=True)

        self.kwh = self.df["kwh"].apply(ast.literal_eval)
        self.kwh = torch.from_numpy(np.array(self.kwh.tolist())).float()
        self.month = self.df["month"]
        self.dayofweek = self.df["dayofweek"]

    def standardise(self, x: torch.tensor) -> torch.tensor:
        """
        Standardise kWh with mean 0 and std 1

        Args:
            x (torch.tensor): Input kWh

        Returns:
            torch.tensor: Standardised kWh
        """
        return (x - self.feature_mean) / self.feature_std

    def reconstruct(self, xhat: torch.tensor) -> torch.tensor:
        """
        Reconstruct kWh from standardised values

        Args:
            xhat (torch.tensor): standardised kWh

        Returns:
            torch.tensor: reconstructed kWh
        """
        return (xhat * self.feature_std) + self.feature_mean

    def __len__(self):
        return len(self.df)

    def __getitem__(self, idx):
        standardised_kwh = self.standardise(self.kwh[idx])
        mth = self.month[idx]
        dow = self.dayofweek[idx]

        return standardised_kwh, mth, dow


class LCLDataModule(pl.LightningDataModule):
    """
    Low Carbon London data module
    """

    def __init__(
        self,
        data_path: Path,
        stats_path: Path,
        batch_size: int,
        n_samples: int,
    ):
        super().__init__()
        self.data_path = data_path
        self.stats_path = stats_path
        self.batch_size = batch_size
        self.n_samples = n_samples

    def prepare_data(self):
        pass

    def setup(self, stage=""):
        self.dataset = LCLData(
            data_path=self.data_path,
            stats_path=self.stats_path,
            n_samples=self.n_samples,
        )

    def train_dataloader(self):
        return DataLoader(
            self.dataset,
            self.batch_size,
            drop_last=True,
            shuffle=False,
            generator=g,
        )

    def reconstruct_kwh(self, xhat: torch.tensor) -> torch.tensor:
        return self.dataset.reconstruct(xhat)



# Copyright Contributors to the Opensynth-energy Project.
# SPDX-License-Identifier: Apache-2.0

from opensynth.models.faraday.model import (
    Decoder,
    Encoder,
    FaradayModel,
    FaradayVAE,
)

__all__ = ["FaradayVAE", "FaradayModel", "Encoder", "Decoder"]



# Copyright Contributors to the Opensynth-energy Project.
# SPDX-License-Identifier: Apache-2.0

import logging
import os
import subprocess
from os import path
from pathlib import Path

import wget


def check_redownload(file_path: Path) -> bool:
    """
    Check if file already exists.
    If it does, ask user if they want to download it again.

    Args:
        file_path (str): file path

    Returns:
        bool: True if user expresses intent
        to download it again, False otherwise
    """
    if path.exists(file_path):
        prompt = f"{file_path} already exists. Redownload? (y/n): "
        return True if input(prompt) == "y" else False
    return True


def download_data(url: str, filename: Path, unzip: bool = False):
    """
    Download data from a URL and save it to a file.
    Function will check if file already exists,
    and if does, prompt user to confirm redownload.

    Args:
        url (str): URL to download the data from
        filename (str): File path to save the downloaded data
    """
    if check_redownload(filename):
        if filename.exists():
            os.remove(filename)
        logging.info(f"Downloading data from: {url}")
        wget.download(url, filename)
    else:
        logging.info("Skipping download")

    if unzip:
        # Note: LCL dataset needs to be unzipped in command line with unzip
        # It's compressed using algorithms that Python's
        # zipfile module can't handle
        subprocess.run(["unzip", filename, "-d", "data/raw"])



# Copyright Contributors to the Opensynth-energy Project.
# SPDX-License-Identifier: Apache-2.0

import logging
from pathlib import Path

from opensynth.datasets import datasets_utils
from opensynth.datasets import low_carbon_london as lcl

LCL_URL = "https://data.london.gov.uk/download/smartmeter-energy-use-data-in-london-households/3527bf39-d93e-4071-8451-df2ade1ea4f2/LCL-FullData.zip"  # noqa
FILE_NAME = Path("data/raw/lcl_full_data.zip")  # noqa
CSV_FILE_NAME = Path("data/raw/CC_LCL-FullData.csv")  # noqa

logger = logging.getLogger(__name__)


def get_lcl_data(download: bool, split: bool, preprocess: bool):
    """
    Download, split and preprocess the Low Carbon London dataset.
    Download=True downloads and decompress data from data.london.gov.uk.
    Split=True splits the LCL households into training and holdout sets.
    Preprocess=True preprocesses the LCL data into daily load profiles.

    Notes:
    - The LCL dataset is a large dataset and may take a while to download.
    - The compressed version of LCL dataset is about 700MB, but the full
    dataset size is about 8GB.
    - Decompression may not work on windows. If so, please manually download
    and unzip the contents into the `data/raw` folder.

    Args:
        download (bool): True to download data from data.london.gov.uk.
        split (bool): True to split LCL households into training/ holdout sets.
        preprocess (bool): True to preprocess data.
    """
    logger.info(
        f"Running get_lcl_data with download={download}, "
        f"split={split}, preprocess={preprocess}."
    )

    if download:
        datasets_utils.download_data(LCL_URL, FILE_NAME)
    if split:
        lcl.split_households.split_lcl_data(CSV_FILE_NAME, 2000)
    if preprocess:
        lcl.preprocess_lcl.preprocess_lcl_data()



# Copyright Contributors to the Opensynth-energy Project.
# SPDX-License-Identifier: Apache-2.0

import logging
import os
import random
from pathlib import Path
from typing import List, Tuple

import pandas as pd

random.seed(42)

logger = logging.getLogger(__name__)


def split_household_ids(
    df: pd.DataFrame,
    id_col: str,
    sample_size: int,
) -> Tuple[List[str], List[str]]:
    """
    Split LCL id into training vs holdout households.

    Args:
        df (pd.DataFrame): LCL dataset
        id_col (str): ID column
        sample_size (int): Number of household ids in each split

    Returns:
        Tuple[List[str], List[str]]: List of training and holdout household ids
    """
    logger.info("Splitting households into train and holdout households")
    unique_ids = df[id_col].unique().tolist()
    random.shuffle(unique_ids)

    train_ids = unique_ids[:sample_size]
    holdout_ids = unique_ids[-sample_size:]

    return train_ids, holdout_ids


def split_historical_future_periods(
    df: pd.DataFrame, date_col: str = "DateTime"
) -> Tuple[pd.DataFrame, pd.DataFrame]:
    """
    Splits LCL dataset into:
     1) historical period (1st Jan 2012 to 31st Dec 2013)
     2) future period (after 1st Jan 2014)

     These periods are used for TSTR evaluation.

    Args:
        df (pd.DataFrame): pd.DataFrame
        date_col (str, optional): Name of date column. Defaults to "DateTime".

    Returns:
        Tuple[pd.DataFrame, pd.DataFrame]: Historical and Future dataframe.
    """
    historical_start = df[date_col] >= "2012-01-01"
    historical_end = df[date_col] <= "2013-12-31"
    tstr_historical_mask = historical_start & historical_end

    future_start = df[date_col] >= "2014-01-01"
    future_end = df[date_col] <= "2014-12-31"
    tstr_future_mask = future_start & future_end

    df_historical = df.loc[tstr_historical_mask]
    df_future = df.loc[tstr_future_mask]
    return df_historical, df_future


def split_lcl_data(csv_filename: Path, sample_size: int = 2000):
    """
    Split LCL dataset 4 ways:
    1) Historical Train household data
    2) Historical Holdout household data
    3) Future Train household data
    4) Future Holdout household data

    Historical data is used for training generative models.
    Future data is used for TSTR evaluation.

    Args:
        sample_size (int, optional): _description_. Defaults to 2000.
    """

    logger.info(f"👀 Reading LCL data from: {csv_filename}")
    df = pd.read_csv(csv_filename)

    logger.info("🖖 Spliting households into train and holdout")
    train_ids, holdout_ids = split_household_ids(
        df,
        id_col="LCLid",
        sample_size=sample_size,
    )
    logger.info(f"Train len: {len(train_ids)}")
    logger.info(f"Holdout len: {len(holdout_ids)}")

    logger.info("📆 Splitting data into train and holdout period")
    df_history, df_future = split_historical_future_periods(df)
    logger.info(f"History len: {len(df_history)}")
    logger.info(f"Future len: {len(df_future)}")

    df_historical_train = df_history[df_history["LCLid"].isin(train_ids)]
    df_future_train = df_future[df_future["LCLid"].isin(train_ids)]
    df_historical_holdout = df_history[df_history["LCLid"].isin(holdout_ids)]
    df_future_holdout = df_future[df_future["LCLid"].isin(holdout_ids)]

    logger.info("📦 Saving train and holdout data")
    historical_path = Path("data/processed/historical")
    future_path = Path("data/processed/future")
    os.makedirs(historical_path, exist_ok=True)
    os.makedirs(future_path, exist_ok=True)
    df_historical_train.to_csv(
        f"{historical_path}/train.csv",
        index=False,
    )
    df_historical_holdout.to_csv(
        f"{historical_path}/holdout.csv",
        index=False,
    )
    df_future_train.to_csv(f"{future_path}/train.csv", index=False)
    df_future_holdout.to_csv(f"{future_path}/holdout.csv", index=False)

    logger.info("👍 Done!")



# Copyright Contributors to the Opensynth-energy Project.
# SPDX-License-Identifier: Apache-2.0

import csv
import logging
import os
from pathlib import Path
from typing import Tuple

import numpy as np
import pandas as pd

logger = logging.getLogger(__name__)


def get_current_month_end(df: pd.DataFrame, date_col="dt"):
    """
    Get the month end of current month.
    Note: Pandas offsets.MonthEnd()
    returns the date of the following month instead of the current month.

    E.g. if dt = 31st Jan, offsets.MonthEnd returns 28th Feb
    instead of 31st Jan. To work around this, we offset by -1 days
    to get previous day, then do offsets.MonthEnd()

    This is faster than offsets.MonthEnd().rollforward(ts):
    - This method doesn't work on array, only works on raw Timestamp object,
    which will require for loops (or .apply) which is super slow on big data
    - More efficient to perform this in a vectorised manner by offset
    with -1 days then use offsets.MonthEnd().

    Args:
        df (Pandas dataframe): Input pandas dataframe
        date_col (str, optional): Name of date column. Defaults to "dt".

    Returns:
        pd.DataFrame: Output dataframe with the "month_end" column
    """
    df["month_end"] = df[date_col] + pd.offsets.Day(-1)
    df["month_end"] = df["month_end"] + pd.offsets.MonthEnd()
    df["month_end"] = df["month_end"].dt.date
    return df


def load_data(file_path: Path) -> pd.DataFrame:
    """
    Load LCL data from csv

    Args:
        files_path (str): Folder containing CSV files

    Returns:
        pd.DataFrame: LCL dataset
    """
    logger.info(f"🚛 Loading data from {file_path}")
    df = pd.read_csv(file_path)
    df = df.rename(columns={"KWH/hh (per half hour) ": "kwh"})
    return df


def extract_date_features(
    df: pd.DataFrame,
) -> pd.DataFrame:
    """
    Parse date features from DateTime column.

    Args:
        df (pd.DataFrame): Input dataframe

    Returns:
        pd.DataFrame: Output dataframe with date features
    """
    logger.info("📅 Extracting date features")
    df["dt"] = pd.to_datetime(df["DateTime"])
    df["date"] = df["dt"].dt.date.astype(str)
    df["month"] = df["dt"].dt.month.astype(int)
    df["week"] = df["dt"].dt.to_period("W-MON").dt.start_time
    df = get_current_month_end(df, date_col="dt")
    df["month_max"] = pd.to_datetime(df["month_end"]).dt.day.astype(int)
    df["day"] = df["dt"].dt.day.astype(int)
    df["dayofweek"] = df["dt"].dt.dayofweek.astype(int)
    df["hour"] = df["dt"].dt.hour.astype(int)
    df["minute"] = df["dt"].dt.minute.astype(int)
    return df


def parse_settlement_period(
    df: pd.DataFrame,
) -> pd.DataFrame:
    """
    Parse settlement periods from hour and minute columns

    Args:
        df (pd.DataFrame): Input dataframe

    Returns:
        pd.DataFrame: Output dataframe with settlement period column
    """
    logger.info("🕰 Parsing Settlement Period")

    def _get_settlement_offset(minute_value):
        if minute_value >= 30:
            return 1
        return 0

    df_out = df.copy()
    df_out["settlement_offset"] = df_out["minute"].apply(
        _get_settlement_offset
    )
    df_out["settlement_period"] = (
        df_out["hour"] * 2 + df_out["settlement_offset"] + 1
    )
    df_out = df_out.drop(columns=["settlement_offset"])

    return df_out


def drop_dupes_and_replace_nulls(df: pd.DataFrame) -> pd.DataFrame:
    """
    Function to drop duplicated readings and replace missing readings with 0.0

    Args:
        df (pd.DataFrame): Input dataframe

    Returns:
        pd.DataFrame: Output dataframe
    """
    logger.info("🗑 Dropping dupes and filling nulls with 0")
    df_out = df.copy()
    df_out = df_out.sort_values(
        by=["LCLid", "date", "settlement_period"], ascending=True
    )
    df_out = df_out.drop_duplicates(
        subset=["LCLid", "date", "settlement_period"], keep="last"
    )
    df_out["kwh"] = df_out["kwh"].replace("Null", np.float64())
    df_out["kwh"] = df_out["kwh"].astype(float)
    return df_out


def filter_missing_kwh(df: pd.DataFrame) -> pd.DataFrame:
    """
    Drop dates where we don't have full 48 readings
    for a given LCLid and date.

    Args:
        df (pd.DataFrame): _description_

    Returns:
        pd.DataFrame: _description_
    """
    logger.info("🔍 Filtering missing kwh readings")
    id_col = ["LCLid"]
    merge_cols = id_col + ["date"]
    df_group = df.groupby(merge_cols)[["kwh"]].count().reset_index()
    df_group["required_len"] = 48  # 48 hh readings

    df_full_data = df_group.query("required_len==kwh")  # Has all required data
    df_out = df_full_data[merge_cols].merge(df, on=merge_cols, how="inner")
    return df_out


def pack_smart_meter_data_into_arrays(df: pd.DataFrame) -> pd.DataFrame:
    """
    Pack smart meter data into 48-dimensional arrays.
    Note: LCL dataset are all given in UTC. We should expect
    48-readings per day.

    Args:
        df (pd.DataFrame): Input dataframe

    Returns:
        pd.DataFrame: Output dataframe
    """
    logger.info("👝 Packing time series into arrays")
    df_out = df.copy()
    df_out = df_out.sort_values(
        by=["LCLid", "date", "settlement_period"], ascending=True
    )
    groupby_cols = ["LCLid", "stdorToU", "month_end", "month"]
    groupby_cols = groupby_cols + ["dayofweek", "day", "date"]

    df_out = pd.DataFrame(
        df_out.groupby(groupby_cols)["kwh"]
        .agg(lambda x: x.tolist())
        .reset_index()
    )
    return df_out


def get_mean_and_std(df: pd.DataFrame) -> Tuple[float, float]:
    """
    Calculate the mean and standard deviation of the dataset

    Args:
        df (pd.DataFrame): Input dataframe

    Returns:
        Tuple[float, float]: Mean and standard deviation
    """
    mean = np.mean(df["kwh"])
    std = np.std(df["kwh"])
    return mean, std


def preprocess_pipeline(file_path: Path, out_path: Path):

    df = load_data(file_path)
    df = extract_date_features(df)
    df = parse_settlement_period(df)
    df = drop_dupes_and_replace_nulls(df)
    df = filter_missing_kwh(df)

    mean, stdev = get_mean_and_std(df)
    df = pack_smart_meter_data_into_arrays(df)

    os.makedirs(out_path, exist_ok=True)
    df.to_csv(f"{out_path}/lcl_data.csv", index=False)

    mean_std_dict = {"mean": mean, "stdev": stdev}
    with open(f"{out_path}/mean_std.csv", "w") as f:
        w = csv.DictWriter(f, mean_std_dict.keys())
        w.writeheader()
        w.writerow(mean_std_dict)


def preprocess_lcl_data():

    SOURCE_DIR = "data/raw"
    OUT_DIR = "data/processed"
    preprocess_pipeline(
        file_path=f"{SOURCE_DIR}/historical/train.csv",
        out_path=f"{OUT_DIR}/historical/train",
    )
    preprocess_pipeline(
        file_path=f"{SOURCE_DIR}/historical/holdout.csv",
        out_path=f"{OUT_DIR}/historical/holdout",
    )
    preprocess_pipeline(
        file_path=f"{SOURCE_DIR}/future/train.csv",
        out_path=f"{OUT_DIR}/future/train",
    )
    preprocess_pipeline(
        file_path=f"{SOURCE_DIR}/future/holdout.csv",
        out_path=f"{OUT_DIR}/future/holdout",
    )


if __name__ == "__main__":
    preprocess_lcl_data()



# Copyright Contributors to the Opensynth-energy Project.
# SPDX-License-Identifier: Apache-2.0

from typing import Tuple

import torch
import torch.nn.functional as F


def MMDLoss(y: torch.tensor, x: torch.tensor) -> torch.tensor:
    """
    Calculate MMD Loss

    Args:
        y (torch.tensor): Tensor Y
        x (torch.tensor): Tensor X

    Returns:
        torch.tensor: MMD Distances between Tensor Y and X
    """

    xx, yy, zz = torch.mm(x, x.t()), torch.mm(y, y.t()), torch.mm(x, y.t())
    rx = xx.diag().unsqueeze(0).expand_as(xx)
    ry = yy.diag().unsqueeze(0).expand_as(yy)

    dxx = rx.t() + rx - 2.0 * xx
    dyy = ry.t() + ry - 2.0 * yy
    dxy = rx.t() + ry - 2.0 * zz

    XX, YY, XY = (
        torch.zeros(xx.shape),
        torch.zeros(xx.shape),
        torch.zeros(xx.shape),
    )
    XX = XX.to(x)
    YY = YY.to(x)
    XY = XY.to(x)

    bandwidth_range = [10, 15, 20, 50]
    for a in bandwidth_range:
        XX += torch.exp(-0.5 * dxx / a)
        YY += torch.exp(-0.5 * dyy / a)
        XY += torch.exp(-0.5 * dxy / a)

    return torch.mean(XX + YY - 2.0 * XY)


def quantile_loss(
    y_pred: torch.tensor, y_real: torch.tensor, quantile: float
) -> torch.tensor:
    """
    Calculate quantile loss
    #TODO: Test this function
    Args:
        y_pred (torch.tensor): Predicted quantile
        y_real (torch.tensor): Actual quantile
        quantile (float): Quantile value

    Returns:
        torch.tensor: Quantile loss
    """
    return torch.mean(
        torch.max(
            quantile * (y_real - y_pred), (1 - quantile) * (y_pred - y_real)
        )
    )


def calculate_training_loss(
    x_hat: torch.tensor,
    x: torch.tensor,
    mse_weight: float,
    quantile_upper_weight: float,
    quantile_lower_weight: float,
    quantile_median_weight: float,
    lower_quantile: float,
    upper_quantile: float,
) -> Tuple[torch.tensor, torch.tensor, torch.tensor, torch.tensor]:
    """
    Calculate training losses for Faraday.
    Losses are a combined total of:
    1) MMD loss
    2) MSE loss * mse_weight
    3) upper quantile loss * upper quantile weight
    4) lower quantile loss * lower quantile weight
    5) median quantile loss * median quantile weight
    Args:
        x_hat (torch.tensor): Generated tensor
        x (torch.tensor): Real tensor
        mse_weight (float): MSE Weight
        quantile_upper_weight (float): Upper quantile weight
        quantile_lower_weight (float): Lower quantile weight
        quantile_median_weight (float): Median quantile weight
        lower_quantile (float): Lower quantile value
        upper_quantile (float): Upper quantile value

    Returns:
        Tuple[torch.tensor, torch.tensor, torch.tensor, torch.tensor]:
        Total loss, MMD loss, MSE loss, Quantile loss
    """
    mmd_loss = MMDLoss(x_hat, x)
    mse_loss = F.mse_loss(x_hat, x) * mse_weight
    quantile_upper_loss = (
        quantile_loss(x_hat, x, upper_quantile) * quantile_upper_weight
    )
    quantile_lower_loss = (
        quantile_loss(x_hat, x, lower_quantile) * quantile_lower_weight
    )
    quantile_median_loss = (
        quantile_loss(x_hat, x, 0.5) * quantile_median_weight
    )
    quantile_losses = (
        quantile_upper_loss + quantile_lower_loss + quantile_median_loss
    )

    total_loss = mmd_loss + mse_loss + quantile_losses

    return total_loss, mmd_loss, mse_loss, quantile_losses



# Copyright Contributors to the Opensynth-energy Project.
# SPDX-License-Identifier: Apache-2.0

import logging
from dataclasses import dataclass
from typing import Optional, Tuple

import numpy as np
import pytorch_lightning as pl
import torch
import torch.nn as nn
from opacus import PrivacyEngine
from opacus.validators import ModuleValidator
from sklearn.mixture import GaussianMixture
from torch.optim import lr_scheduler
from tqdm import tqdm

from opensynth.data_modules.lcl_data_module import LCLDataModule
from opensynth.models.faraday.losses import calculate_training_loss

logger = logging.getLogger(__name__)


@dataclass
class TrainingData:
    kwh: torch.tensor
    month: torch.tensor
    dow: torch.tensor


class Encoder(nn.Module):
    def __init__(self, latent_dim: int, input_dim: int, class_dim: int):
        super().__init__()
        self.latent_dim = latent_dim
        self.input_dim = input_dim
        self.class_dim = class_dim
        self.encoder_input_dim = self.input_dim + self.class_dim

        # Encoder layers
        self.encoder_layers = nn.Sequential(
            nn.Linear(self.encoder_input_dim, 512),
            nn.GELU(),
            nn.Linear(512, 256),
            nn.GELU(),
            nn.Linear(256, 128),
            nn.GELU(),
            nn.Linear(128, 64),
            nn.GELU(),
            nn.Linear(64, 32),
            nn.GELU(),
            nn.Linear(32, self.latent_dim),
        )

    def forward(self, x):
        return self.encoder_layers(x)


class Decoder(nn.Module):
    def __init__(self, class_dim: int, latent_dim: int, output_dim: int):
        super().__init__()
        self.latent_dim = latent_dim
        self.class_dim = class_dim
        self.output_dim = output_dim
        self.decoder_input_dim = self.latent_dim + self.class_dim

        # Layers to map latent space back to FC layers
        self.latent = nn.Linear(self.decoder_input_dim, self.latent_dim)
        self.latent_activations = nn.GELU()

        # Decoder layers
        self.decoder_layers = nn.Sequential(
            nn.Linear(self.latent_dim, 32),
            nn.GELU(),
            nn.Linear(32, 64),
            nn.GELU(),
            nn.Linear(64, 128),
            nn.GELU(),
            nn.Linear(128, 256),
            nn.GELU(),
            nn.Linear(256, 512),
            nn.GELU(),
            nn.Linear(512, self.output_dim),
        )

    def forward(self, x):
        outputs = self.latent(x)
        outputs = self.latent_activations(outputs)
        outputs = self.decoder_layers(outputs)
        return outputs


class ReparametrisationModule(nn.Module):
    def __init__(self, latent_dim: int):
        super().__init__()
        self.latent_dim = latent_dim
        self.mean = nn.Linear(self.latent_dim, self.latent_dim)
        self.logvar = nn.Linear(self.latent_dim, self.latent_dim)
        self.norm_dist = torch.distributions.Normal(0, 1)

    def forward(self, encoded_x):
        mu = self.mean(encoded_x)
        sigma = self.logvar(encoded_x)
        eps = self.norm_dist.sample(mu.shape)
        eps = eps.to(encoded_x)
        return mu + eps * sigma, mu, sigma


class FaradayVAE(pl.LightningModule):
    def __init__(
        self,
        class_dim: int,
        learning_rate: float = 1e-3,
        latent_dim: int = 16,
        input_dim: int = 48,
        tmax: int = 5000,
        mse_weight: float = 2.0,
        quantile_upper_weight: float = 4,
        quantile_lower_weight: float = 1,
        quantile_median_weight: float = 4,
        lower_quantile: float = 0.05,
        upper_quantile: float = 0.95,
        differential_privacy: bool = False,
        epsilon: Optional[float] = 1.0,
        delta: Optional[float] = 1e-5,
        max_grad_norm: Optional[float] = 1.0,
        custom_encoder: Optional[Encoder] = None,
        custom_decoder: Optional[Decoder] = None,
    ):
        super().__init__()
        self.class_dim = class_dim
        self.latent_dim = latent_dim
        self.input_dim = input_dim
        self.tmax = tmax
        self.learning_rate = learning_rate
        self.mse_weight = mse_weight
        self.quantile_upper_weight = quantile_upper_weight
        self.quantile_lower_weight = quantile_lower_weight
        self.quantile_median_weight = quantile_median_weight
        self.lower_quantile = lower_quantile
        self.upper_quantile = upper_quantile
        self.differential_privacy = differential_privacy

        if self.differential_privacy:

            if max_grad_norm is None:
                raise ValueError("Max grad norm must be set for DP training")
            elif max_grad_norm > 1 or max_grad_norm < 0:
                raise ValueError("Max grad norm must be between 0 and 1")

            logger.info("🔒 Differential Privacy Enabled")
            logger.info("🔒 Epsilon: {epsilon}, Delta: {delta}")
            logger.info(
                "🔒 Note: to satisfy definition of"
                "differential privacy, Delta: {delta}"
                "must be < 1/N where N is the size of"
                "the training dataset"
            )
            self.max_grad_norm = max_grad_norm
            self.epsilon = epsilon
            self.delta = delta

            self.privacy_engine = PrivacyEngine(secure_mode=False)
            # Check that everything is valid for DP training
            assert ModuleValidator.validate(self, strict=True) == []

        # Save hyperparameters
        self.save_hyperparameters(ignore=["custom_encoder", "custom_decoder"])

        self.encoder = (
            custom_encoder
            if custom_encoder is not None
            else Encoder(self.latent_dim, self.input_dim, self.class_dim)
        )
        self.decoder = (
            custom_decoder
            if custom_decoder is not None
            else Decoder(self.class_dim, self.latent_dim, self.input_dim)
        )
        self.reparametriser = ReparametrisationModule(self.latent_dim)

    def encode(self, input_tensor: torch.tensor):
        encoded_x = self.encoder(input_tensor)
        encoded_x, _, _ = self.reparametriser(encoded_x)
        return encoded_x

    def decode(self, latent_tensor: torch.tensor):
        return self.decoder(latent_tensor)

    def configure_optimizers(self):

        optim = torch.optim.Adam(self.parameters(), lr=self.learning_rate)

        if self.differential_privacy:
            self.trainer.fit_loop.setup_data()
            dataloader = self.trainer.train_dataloader
            epochs = self.trainer.max_epochs
            model, optim, dl = self.privacy_engine.make_private_with_epsilon(
                module=self,
                optimizer=optim,
                data_loader=dataloader,
                epochs=epochs,
                target_epsilon=self.epsilon,
                target_delta=self.delta,
                max_grad_norm=self.max_grad_norm,
            )
            self.dp = {"model": model, "optim": optim, "dataloader": dl}

            return optim

        else:
            # Learning Rate schedulers are not compatible with
            # Opacus and Pytorch Lightning
            lr_schedule = lr_scheduler.CosineAnnealingLR(
                optim, T_max=self.tmax, eta_min=self.learning_rate / 5
            )
            return [optim], [lr_schedule]

    def forward(self, input_data: TrainingData):
        kwh = input_data.kwh
        month = input_data.month.reshape(len(kwh), 1)
        dow = input_data.dow.reshape(len(kwh), 1)

        encoder_inputs = torch.cat([kwh, month, dow], dim=1)
        encoder_outputs = self.encode(encoder_inputs)
        decoder_inputs = torch.cat([encoder_outputs, month, dow], dim=1)
        decoder_outputs = self.decode(decoder_inputs)

        return decoder_outputs

    def training_step(self, batch):
        batch_data = TrainingData(kwh=batch[0], month=batch[1], dow=batch[2])
        vae_outputs = self.forward(batch_data)
        total_loss, mmd_loss, mse_loss, quantile_loss = (
            calculate_training_loss(
                x_hat=vae_outputs,
                x=batch_data.kwh,
                lower_quantile=self.lower_quantile,
                upper_quantile=self.upper_quantile,
                mse_weight=self.mse_weight,
                quantile_upper_weight=self.quantile_upper_weight,
                quantile_lower_weight=self.quantile_lower_weight,
                quantile_median_weight=self.quantile_median_weight,
            )
        )

        # Might be an overkill to sync_dist for all losses.
        # If this causes significant I/O bottleneck
        # Consider syncing only total_loss
        self.log(
            "total_loss",
            total_loss,
            on_epoch=True,
            on_step=False,
            prog_bar=True,
            sync_dist=True,
        )
        self.log(
            "mmd_loss",
            mmd_loss,
            on_epoch=True,
            on_step=False,
            prog_bar=True,
            sync_dist=True,
        )
        self.log(
            "mse_loss",
            mse_loss,
            on_epoch=True,
            on_step=False,
            prog_bar=True,
            sync_dist=True,
        )
        self.log(
            "quantile_loss",
            quantile_loss,
            on_epoch=True,
            on_step=False,
            prog_bar=True,
            sync_dist=True,
        )
        return total_loss

    def on_train_epoch_end(self):
        if self.differential_privacy:
            # Note: Need to convert to float32 because MPS
            # device does not support float64. Will error if you don't convert
            eps = np.float32(self.privacy_engine.get_epsilon(self.delta))
            self.log(
                "epsilon",
                eps,
                on_epoch=True,
                on_step=False,
                prog_bar=True,
                sync_dist=False,
            )


class FaradayModel:
    def __init__(
        self,
        vae_module: FaradayVAE,
        n_components: int,
        max_iter: int = 1000,
        covariance_type: str = "full",
        tol: float = 1e-3,
    ):
        """
        Faraday Model

        Args:
            vae_module (FaradayVAE): Trained VAE component
            n_components (int): GMM clusteres
            max_iter (int, optional): Max iteration for GMM. Defaults to 1000.
            covariance_type (str, optional): scikit-learn gmm covariance types.
                Defaults to "full".
            tol (float, optional): Tolerance for GMM. Defaults to 1e-3.
        """
        self.n_components = n_components
        self.max_iter = max_iter
        self.covariance_type = covariance_type
        self.vae_module = vae_module
        self.tol = tol

        self.gmm = GaussianMixture(
            n_components=n_components,
            max_iter=max_iter,
            covariance_type=covariance_type,
            warm_start=True,
            tol=tol,
        )

    def train_gmm(self, dm: LCLDataModule):
        """
        Train Gaussian Mixture Module

        Args:
            dm (LCLDataModule): Training data
        """
        dl = dm.train_dataloader()
        for batch_num, batch_data in tqdm(enumerate(dl)):
            kwh = batch_data[0]
            mth = batch_data[1].reshape(len(kwh), 1)
            dow = batch_data[2].reshape(len(kwh), 1)
            vae_input = torch.cat([kwh, mth, dow], dim=1)
            vae_output = self.vae_module.encode(vae_input)
            gmm_input = torch.cat([vae_output, mth, dow], dim=1)
            self.gmm.fit(gmm_input.detach().numpy())
            logger.info(f"⏳ Batch {batch_num} completed")

        self.max_mth = mth.max().item()
        self.min_mth = mth.min().item()
        self.max_dow = dow.max().item()
        self.min_dow = dow.min().item()
        logger.info(
            f"Labels Min max: Max month: {self.max_mth},"
            "Min month: {self.min_mth}"
            ", Max dow: {self.max_dow}, Min dow: {self.min_dow}"
        )
        logger.info("🎉 GMM Training Completed")

    def sample_gmm(
        self, n_samples: int
    ) -> Tuple[torch.tensor, torch.tensor, torch.tensor]:
        """
        Samples latent codes from GMM and decode with decoder.

        Args:
            n_samples (int): Number of samples to generate.

        Returns:
            Tuple[torch.tensor, torch.tensor, torch.tensor]:
              Decoder output (KWH), month label, dow label
        """
        gmm_samples = self.gmm.sample(n_samples)[0]
        gmm_kwh = gmm_samples[:, : self.vae_module.latent_dim]
        gmm_mth = np.round(gmm_samples[:, -2], decimals=0).astype(int)
        gmm_dow = np.round(gmm_samples[:, -1], decimals=0).astype(int)

        # Filter invalid (out of distribution) samples
        label_mask = (
            (gmm_mth >= self.min_mth)
            & (gmm_mth <= self.max_mth)
            & (gmm_dow >= self.min_dow)
            & (gmm_dow <= self.max_dow)
        )
        gmm_kwh = gmm_kwh[label_mask]
        gmm_mth = gmm_mth[label_mask]
        gmm_dow = gmm_dow[label_mask]

        latent_tensor = torch.from_numpy(gmm_kwh)
        mth_tensor = torch.from_numpy(gmm_mth).reshape(len(latent_tensor), 1)
        dow_tensor = torch.from_numpy(gmm_dow).reshape(len(latent_tensor), 1)
        decoder_input = torch.cat(
            [latent_tensor, mth_tensor, dow_tensor], dim=1
        ).float()
        decoder_output = self.vae_module.decode(decoder_input)
        return decoder_output, mth_tensor, dow_tensor
